"""Unit-tests for Santec TSL-570 driver class."""
import logging
import random
import unittest
from unittest.mock import call, patch, Mock
import numpy as np

import qmi
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport

from qmi.instruments.santec.tsl_570 import Santec_Tsl570


class TestSantecTsl570OpenClose(unittest.TestCase):

    def setUp(self):
        logging.getLogger("qmi.instruments.santec.tsl_570").setLevel(logging.CRITICAL)
        qmi.start("TestSantecTsl570ClassContext")
        # Add patches
        patcher = patch('qmi.instruments.santec.tsl_570.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start()
        self.addCleanup(patcher.stop)
        patcher2 = patch('qmi.instruments.santec.tsl_570.ScpiProtocol', autospec=ScpiProtocol)
        self._scpi_mock = patcher2.start()
        self.addCleanup(patcher2.stop)
        self._scpi_mock.write = Mock()
        # Make DUT
        self.instr: Santec_Tsl570 = Santec_Tsl570(qmi.context(), "Santec_laser", "")

    def tearDown(self):
        qmi.stop()
        logging.getLogger("qmi.instruments.santec.tsl_570").setLevel(logging.NOTSET)

    def test_open_close(self):
        """open() command not just opens the instrument, but immediately also checks the wavelength/frequency
        range and the power unit to set the power level range. Test the whole shebang.

        close() command then just closes the connection to the instrument and the transport.
        """
        # Initially, the wavelength and power range should be bare data classes
        expected_initial_out = "<class 'qmi.instruments.santec.tsl_570._{rng}Range'>"
        self.assertEqual(expected_initial_out.format(rng="Wavelength"), str(self.instr._wavelength_range))
        self.assertEqual(expected_initial_out.format(rng="Frequency"), str(self.instr._frequency_range))
        self.assertEqual(expected_initial_out.format(rng="PowerLevel"), str(self.instr._power_level_range))
        # Arrange
        expected_wl_min = 1200.0
        expected_wl_max = 1700.0
        expected_freq_min = 175.0
        expected_freq_max = 250.0
        power_unit = 0  # "dBm"
        expected_power_min = -15.0
        expected_power_max = 13.0

        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_wl_min}\r",
            f"{expected_wl_max}\r",
            f"{expected_freq_min}\r",
            f"{expected_freq_max}\r",
            f"{power_unit}\r"
        ]
        expected_calls = [
            call().ask(":WAV:MIN?"),
            call().ask(":WAV:MAX?"),
            call().ask(":WAV:FREQ:MIN?"),
            call().ask(":WAV:FREQ:MAX?"),
            call().ask(":POW:UNIT?")
        ]
        # Act
        self.instr.open()
        # Assert
        self._transport_mock.assert_called_once()
        self.assertEqual(expected_wl_min, self.instr._wavelength_range.min)
        self.assertEqual(expected_wl_max, self.instr._wavelength_range.max)
        self.assertEqual(expected_power_min, self.instr._power_level_range.min)
        self.assertEqual(expected_power_max, self.instr._power_level_range.max)
        self._scpi_mock.assert_has_calls(expected_calls, any_order=True)
        # Test close
        self.instr.close()
        self._transport_mock.assert_called_once()

    def test_open_close_alternate_unit(self):
        """Test the open() - close() but with in milliwatt units"""
        # Arrange
        expected_wl_min = 1200.0
        expected_wl_max = 1700.0
        expected_freq_min = 175.0
        expected_freq_max = 250.0
        power_unit = 1  # "mW"
        expected_power_min = 0.04
        expected_power_max = 20.0
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_wl_min}\r",
            f"{expected_wl_max}\r",
            f"{expected_freq_min}\r",
            f"{expected_freq_max}\r",
            f"{power_unit}\r"
        ]
        expected_calls = [
            call().ask(":WAV:MIN?"),
            call().ask(":WAV:MAX?"),
            call().ask(":WAV:FREQ:MIN?"),
            call().ask(":WAV:FREQ:MAX?"),
            call().ask(":POW:UNIT?")
        ]
        # Act
        self.instr.open()
        # Assert
        self._transport_mock.assert_called_once()
        self.assertEqual(expected_wl_min, self.instr._wavelength_range.min)
        self.assertEqual(expected_wl_max, self.instr._wavelength_range.max)
        self.assertEqual(expected_freq_min, self.instr._frequency_range.min)
        self.assertEqual(expected_freq_max, self.instr._frequency_range.max)
        self.assertEqual(expected_power_min, self.instr._power_level_range.min)
        self.assertEqual(expected_power_max, self.instr._power_level_range.max)
        self._scpi_mock.assert_has_calls(expected_calls)
        # Test close
        self.instr.close()
        self._transport_mock.assert_called_once()


class TestSantecTsl570ClassMethods(unittest.TestCase):

    def setUp(self):
        # Make as standard to open in wavelength and dBm mode.
        self.wl_min = 1200.0
        self.wl_max = 1700.0
        self.freq_min = 175.0
        self.freq_max = 250.0
        power_unit = 0  # "dBm"
        self.power_min = -15.0
        self.power_max = 13.0

        logging.getLogger("qmi.instruments.santec.tsl_570").setLevel(logging.CRITICAL)
        qmi.start("TestSantecTsl570ClassContext")
        # Add patches
        patcher = patch('qmi.instruments.santec.tsl_570.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start()
        self.addCleanup(patcher.stop)
        patcher2 = patch('qmi.instruments.santec.tsl_570.ScpiProtocol', autospec=ScpiProtocol)
        self._scpi_mock = patcher2.start()
        self.addCleanup(patcher2.stop)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.wl_min}\r",
            f"{self.wl_max}\r",
            f"{self.freq_min}\r",
            f"{self.freq_max}\r",
            f"{power_unit}\r"
        ]
        self._scpi_mock.write = Mock()
        # Make DUT
        self.instr: Santec_Tsl570 = Santec_Tsl570(qmi.context(), "RSB100a", "")
        self.instr.open()
        # Reset the SCPI mock for further tests
        self._scpi_mock.reset_mock()

    def tearDown(self):
        self.instr.close()
        qmi.stop()
        logging.getLogger("qmi.instruments.santec.tsl_570").setLevel(logging.NOTSET)

    def test_wrong_int_value(self):
        """Test a call that uses _ask_int to see that it raises an error at wrong response"""
        self._scpi_mock(self._transport_mock).ask.side_effect = ["not an int"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_coherence_control_status()

    def test_wrong_float_value(self):
        """Test a call that uses _ask_float to see that it raises an error at wrong response"""
        self._scpi_mock(self._transport_mock).ask.side_effect = ["not a float"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_wavelength()

    def test_wrong_bool_value(self):
        """Test a call that uses _ask_bool to see that it raises an error at wrong response"""
        self._scpi_mock(self._transport_mock).ask.side_effect = ["not a bool"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.operation_complete()

    def test_get_idn(self):
        """Test ident. """
        vendor = "vendor"
        model = "model"
        serial = "serial"
        version = "version"
        self._scpi_mock(self._transport_mock).ask.side_effect = [f"{vendor},{model},{serial},{version}"]

        ident = self.instr.get_idn()

        self._scpi_mock.assert_has_calls([call().ask("*IDN?")])
        self.assertEqual(ident.vendor, vendor)
        self.assertEqual(ident.model, model)
        self.assertEqual(ident.serial, serial)
        self.assertEqual(ident.version, version)

    def test_wrong_idn_response(self):
        """Test ident. raises QMI_InstrumentException by a non-sense response."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ["nonsense"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()

        self._scpi_mock.assert_has_calls([call().ask("*IDN?")])

    def test_reset(self):
        """Test reset."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ['0,"No error"', "No alerts."]

        self.instr.reset()

        self._scpi_mock.assert_has_calls([call().write("*RST")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ERR?")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ALER?")])

    def test_clear(self):
        """Test clear."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ['0,"No error"', "No alerts."]

        self.instr.clear()

        self._scpi_mock.assert_has_calls([call().write("*CLS")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ERR?")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ALER?")])

    def test_operation_completed(self):
        """Test getting OPC status."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ["1"]

        opc = self.instr.operation_complete()

        self.assertTrue(opc)
        self._scpi_mock.assert_has_calls([call().ask("*OPC?")])

    def test_error(self):
        """Test command with error and alert responses. '0,"No error"' should be returned when no further
        errors or alerts are present."""
        # Arrange
        error_code = -410
        alert_code = "No27."
        expected_error = [Santec_Tsl570.ERRORS_TABLE[error_code]]
        expected_alert = [Santec_Tsl570.ALERTS_TABLE[alert_code]]
        expected_exception = f"Command *RST resulted in errors: {expected_error + expected_alert}"
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{error_code}",
            '0,"No error"',
            f"{alert_code}",
            "No alerts."
        ]
        expected_write_calls = [
            call().write("*RST")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.reset()
        # Assert
        self.assertEqual(expected_exception, str(exc.exception))
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_errors(self):
        """Test the get_errors returns list of errors."""
        error_code = -102
        alert_code = "No07."
        expected_error = [Santec_Tsl570.ERRORS_TABLE[error_code]]
        expected_alert = [Santec_Tsl570.ALERTS_TABLE[alert_code]]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{error_code}",
            '0,"No error"',
            f"{alert_code}",
            "No alerts."
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        response = self.instr.get_errors()
        # Assert
        self.assertEqual(expected_error + expected_alert, response)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_wavelength_in_nm(self):
        """Test setting the wavelength in nm."""
        input_wl = 1450
        expected_dec = 4  # resolution is 0.1pm, so we should use 4 decimals
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV {input_wl:.{expected_dec}f}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_wavelength(input_wl)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_wavelength_in_nm_excepts(self):
        """Test setting the wavelength in nm with values out-of-bounds."""
        input_wl = [self.wl_min - 1, self.wl_max + 1]
        # Act
        for inp in input_wl:
            with self.assertRaises(ValueError):
                self.instr.set_wavelength(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_wavelength_in_nm(self):
        """Test getting the wavelength in nm."""
        wl = 1450.0
        expected_wl = round(wl, 4)  # We use 4 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{wl}"
        ]
        expected_ask_calls = [
            call().ask(":WAV?"),
        ]
        # Act
        wavelength = self.instr.get_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_minimum_wavelength_in_nm(self):
        """Test getting the minimum wavelength in nm."""
        expected_wl = round(self.wl_min, 4)  # We use 4 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.wl_min}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:MIN?"),
        ]
        # Act
        wavelength = self.instr.get_minimum_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_maximum_wavelength_in_nm(self):
        """Test getting the maximum wavelength in nm."""
        expected_wl = round(self.wl_max, 4)  # We use 4 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.wl_max}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:MAX?"),
        ]
        # Act
        wavelength = self.instr.get_maximum_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_frequency_in_thz(self):
        """Test setting the frequency in thz."""
        input_freq = 231.0
        expected_dec = 5  # resolution is 10MHz, so we should use 5 decimals
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:FREQ {input_freq:.{expected_dec}f}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_frequency(input_freq)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_frequency_in_thz_excepts(self):
        """Test setting the frequency in thz with values out-of-bounds."""
        input_freq = [self.freq_min - 1, self.freq_max + 1]
        # Act
        for inp in input_freq:
            with self.assertRaises(ValueError):
                self.instr.set_frequency(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_frequency_in_thz(self):
        """Test getting the frequency in THz."""
        freq = 200.0
        expected_freq = round(freq, 5)  # We use 5 decimals for rounding output freq.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{freq}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:FREQ?"),
        ]
        # Act
        frequency = self.instr.get_frequency()
        # Assert
        self.assertEqual(expected_freq, frequency)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_minimum_frequency_in_thz(self):
        """Test getting the minimum frequency in THz."""
        expected_freq = round(self.freq_min, 5)  # We use 5 decimals for rounding output freq.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.freq_min}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:FREQ:MIN?"),
        ]
        # Act
        frequency = self.instr.get_minimum_frequency()
        # Assert
        self.assertEqual(expected_freq, frequency)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_maximum_frequency_in_thz(self):
        """Test getting the maximum frequency in THz."""
        expected_freq = round(self.freq_max, 5)  # We use 5 decimals for rounding output freq.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.freq_max}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:FREQ:MAX?"),
        ]
        # Act
        frequency = self.instr.get_maximum_frequency()
        # Assert
        self.assertEqual(expected_freq, frequency)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_wavelength_unit(self):
        """Test setting the wavelength unit."""
        expected_units = [1, 0]  # is ["THz", "nm"]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts.",
            '0,"No error"',
            "No alerts.",
        ]
        expected_calls = [
            call().write(f":WAV:UNIT {expected_units[0]}"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
            call().write(f":WAV:UNIT {expected_units[1]}"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
        ]
        # Act
        self.instr.set_wavelength_unit_to_thz()
        self.instr.set_wavelength_unit_to_nm()

        # Assert
        self._scpi_mock.assert_has_calls(expected_calls)

    def test_get_wavelength_unit_nm(self):
        """Test getting the wavelength unit as nm."""
        expected_unit = "nm"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["0"]
        expected_ask_calls = [
            call().ask(":WAV:UNIT?"),
        ]
        # Act
        unit = self.instr.get_wavelength_unit()
        # Assert
        self.assertEqual(expected_unit, unit)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_wavelength_unit_thz(self):
        """Test getting the wavelength unit as THz."""
        expected_unit = "THz"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["1"]
        expected_ask_calls = [
            call().ask(":WAV:UNIT?"),
        ]
        # Act
        unit = self.instr.get_wavelength_unit()
        # Assert
        self.assertEqual(expected_unit, unit)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_wavelength_fine(self):
        """Test setting the wavelength fine-tuning value."""
        input_wl_fine = 55.0
        expected_dec = 2  # resolution is 0.01, so we should use 2 decimals
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:FIN {input_wl_fine:.{expected_dec}f}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_wavelength_fine(input_wl_fine)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_wavelength_fine_excepts(self):
        """Test setting the wavelength fine-tuning value out-of-bounds."""
        input_wl_fine = [-100.1, 100.1]  # Range is [-100.0, 100.0]
        # Act
        for inp in input_wl_fine:
            with self.assertRaises(ValueError):
                self.instr.set_wavelength_fine(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_wavelength_fine(self):
        """Test getting the wavelength fine-tuning value."""
        wl = 1450.0
        expected_wl = round(wl, 5)  # We use 5 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{wl}"
        ]
        expected_ask_calls = [
            call().ask(":WAV?"),
        ]
        # Act
        wavelength = self.instr.get_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_disable_finetuning_operation(self):
        """Test disabling the fine-tuning operation."""
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts.",
        ]
        expected_write_calls = [
            call().write(":WAV:FIN:DIS"),
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.disable_finetuning_operation()

        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls, any_order=True)
        self._scpi_mock.assert_has_calls(expected_ask_calls, any_order=True)

    def test_set_coherence_control_status(self):
        """Test setting the coherence control_status."""
        inputs = [True, "OFF"]
        expected_control_status = [1, 0]  # is ["ON", "OFF"]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts.",
            '0,"No error"',
            "No alerts.",
        ]
        expected_write_calls = [
            call().write(f":COHC {expected_control_status[0]}"),
            call().write(f":COHC {expected_control_status[1]}"),
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
        ]
        # Act
        for e, inp in enumerate(inputs):
            self.instr.set_coherence_control_status(inp)

        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls, any_order=True)
        self._scpi_mock.assert_has_calls(expected_ask_calls, any_order=True)

    def test_set_coherence_control_status_excepts(self):
        """See that the function excepts if invalid string is given as input."""
        with self.assertRaises(AssertionError):
            self.instr.set_coherence_control_status("AAN")

    def test_get_coherence_control_status(self):
        """Test getting the coherence control_status."""
        expected_control_status = "OFF"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["0"]
        expected_ask_calls = [
            call().ask(":COHC?"),
        ]
        # Act
        control_status = self.instr.get_coherence_control_status()
        # Assert
        self.assertEqual(expected_control_status, control_status)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_optical_output_status(self):
        """Test setting the optical output_status."""
        inputs = [True, "OFF"]
        expected_output_status = [1, 0]  # is ["ON", "OFF"]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts.",
            '0,"No error"',
            "No alerts.",
        ]
        expected_write_calls = [
            call().write(f":POW:STAT {expected_output_status[0]}"),
            call().write(f":POW:STAT {expected_output_status[1]}"),
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
        ]
        # Act
        for e, inp in enumerate(inputs):
            self.instr.set_optical_output_status(inp)

        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls, any_order=True)
        self._scpi_mock.assert_has_calls(expected_ask_calls, any_order=True)

    def test_set_optical_output_status_excepts(self):
        """See that the function excepts if invalid string is given as input."""
        with self.assertRaises(AssertionError):
            self.instr.set_optical_output_status("UIT")

    def test_get_optical_output_status(self):
        """Test getting the optical output_status."""
        expected_output_status = "ON"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["1"]
        expected_ask_calls = [
            call().ask(":POW:STAT?"),
        ]
        # Act
        output_status = self.instr.get_optical_output_status()
        # Assert
        self.assertEqual(expected_output_status, output_status)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_power_level_in_dbm(self):
        """Test setting the power_level in dbm."""
        input_pow = 2
        expected_dec = 2  # resolution is 0.01dbm, so we should use 2 decimals
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            "0",
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":POW {input_pow:.{expected_dec}f}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_power_level(input_pow)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_power_level_in_dbm_excepts(self):
        """Test setting the power_level in dbm with values out-of-bounds."""
        input_pow = [self.power_min - 1, self.power_max + 1]
        self._scpi_mock(self._transport_mock).ask.side_effect = ["0", "0"]
        # Act
        for inp in input_pow:
            with self.assertRaises(ValueError):
                self.instr.set_power_level(inp)

        # Assert
        self._scpi_mock.assert_has_calls([call().ask(":POW:UNIT?")])

    def test_get_power_level_in_dbm(self):
        """Test getting the power_level in dbm."""
        pow = 9.0
        expected_pow = round(pow, 2)  # We use 2 decimals for rounding output pow.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{pow}"
        ]
        expected_ask_calls = [
            call().ask(":POW?"),
        ]
        # Act
        power_level = self.instr.get_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_minimum_power_level_in_dbm(self):
        """Test getting the minimum power_level in dbm."""
        # Arrange
        expected_pow = round(self.power_min, 2)  # We use 2 decimals for rounding output pow.
        # Act
        power_level = self.instr.get_minimum_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_not_called()

    def test_get_maximum_power_level_in_dbm(self):
        """Test getting the maximum power_level in dbm."""
        # Arrange
        expected_pow = round(self.power_max, 5)  # We use 5 decimals for rounding output pow.
        # Act
        power_level = self.instr.get_maximum_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_not_called()

    def test_get_power_level_in_mw(self):
        """Test getting the power_level in mW."""
        pow = 2.0
        expected_pow = round(pow, 2)  # We use 2 decimals for rounding output pow.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{pow}"
        ]
        expected_ask_calls = [
            call().ask(":POW?"),
        ]
        # Act
        power_level = self.instr.get_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_minimum_power_level_in_mw(self):
        """Test getting the minimum power_level in mW."""
        expected_pow = round(self.power_min, 2)  # We use 2 decimals for rounding output pow.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.power_min}"
        ]
        # Act
        power_level = self.instr.get_minimum_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_called_once()

    def test_get_maximum_power_level_in_mw(self):
        """Test getting the maximum power_level in mW."""
        expected_pow = round(self.power_max, 5)  # We use 5 decimals for rounding output pow.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{self.power_max}"
        ]
        # Act
        power_level = self.instr.get_maximum_power_level()
        # Assert
        self.assertEqual(expected_pow, power_level)
        self._scpi_mock.assert_called_once()

    def test_set_power_level_unit(self):
        """Test setting the power_level unit."""
        expected_units = [1, 0]  # is ["mW", "dbm"]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts.",
            '0,"No error"',
            "No alerts.",
        ]
        expected_calls = [
            call().write(f":POW:UNIT {expected_units[0]}"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
            call().write(f":POW:UNIT {expected_units[1]}"),
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?"),
        ]
        # Act
        self.instr.set_power_level_unit_to_mw()
        self.instr.set_power_level_unit_to_dbm()

        # Assert
        self._scpi_mock.assert_has_calls(expected_calls)

    def test_get_power_level_unit_dbm(self):
        """Test getting the power_level unit as dbm."""
        expected_unit = "dBm"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["0"]
        expected_ask_calls = [
            call().ask(":POW:UNIT?"),
        ]
        # Act
        unit = self.instr.get_power_level_unit()
        # Assert
        self.assertEqual(expected_unit, unit)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_power_level_unit_mw(self):
        """Test getting the power_level unit as mW."""
        expected_unit = "mW"
        self._scpi_mock(self._transport_mock).ask.side_effect = ["1"]
        expected_ask_calls = [
            call().ask(":POW:UNIT?"),
        ]
        # Act
        unit = self.instr.get_power_level_unit()
        # Assert
        self.assertEqual(expected_unit, unit)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_start_wavelength_in_nm(self):
        """Test setting the sweep_start_wavelength in nm."""
        input_wl = 1450
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:STAR {input_wl}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_start_wavelength(input_wl)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_start_wavelength_in_nm_excepts(self):
        """Test setting the sweep_start_wavelength in nm with values out-of-bounds."""
        input_wl = [self.wl_min - 1, self.wl_max + 1]
        # Act
        for inp in input_wl:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_start_wavelength(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_start_wavelength_in_nm(self):
        """Test getting the sweep_start_wavelength in nm."""
        wl = 1450.0
        expected_wl = round(wl, 4)  # We use 4 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{wl}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:STAR?"),
        ]
        # Act
        wavelength = self.instr.get_sweep_start_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_stop_wavelength_in_nm(self):
        """Test setting the sweep_stop_wavelength in nm."""
        input_wl = 1450
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:STOP {input_wl}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_stop_wavelength(input_wl)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_stop_wavelength_in_nm_excepts(self):
        """Test setting the sweep_stop_wavelength in nm with values out-of-bounds."""
        input_wl = [self.wl_min - 1, self.wl_max + 1]
        # Act
        for inp in input_wl:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_stop_wavelength(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_stop_wavelength_in_nm(self):
        """Test getting the sweep_stop_wavelength in nm."""
        wl = 1450.0
        expected_wl = round(wl, 4)  # We use 4 decimals for rounding output wl.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{wl}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:STOP?"),
        ]
        # Act
        wavelength = self.instr.get_sweep_stop_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_start_frequency_in_thz(self):
        """Test setting the sweep_start_frequency in thz."""
        input_freq = 201
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:FREQ:SWE:STAR {input_freq}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_start_frequency(input_freq)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_start_frequency_in_thz_excepts(self):
        """Test setting the sweep_start_frequency in thz with values out-of-bounds."""
        input_freq = [self.freq_min - 1, self.freq_max + 1]
        # Act
        for inp in input_freq:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_start_frequency(inp)

        # Assert

        self._scpi_mock.assert_not_called()

    def test_get_sweep_start_frequency_in_thz(self):
        """Test getting the sweep_start_frequency in thz."""
        freq = 1450.0
        expected_freq = round(freq, 5)  # We use 5 decimals for rounding output freq.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{freq}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:FREQ:SWE:STAR?"),
        ]
        # Act
        frequency = self.instr.get_sweep_start_frequency()
        # Assert
        self.assertEqual(expected_freq, frequency)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_stop_frequency_in_thz(self):
        """Test setting the sweep_stop_frequency in thz."""
        input_freq = 201
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:FREQ:SWE:STOP {input_freq}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_stop_frequency(input_freq)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_stop_frequency_in_thz_excepts(self):
        """Test setting the sweep_stop_frequency in thz with values out-of-bounds."""
        input_freq = [self.freq_min - 1, self.freq_max + 1]
        # Act
        for inp in input_freq:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_stop_frequency(inp)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_stop_frequency_in_thz(self):
        """Test getting the sweep_stop_frequency in thz."""
        freq = 1450.0
        expected_freq = round(freq, 5)  # We use 5 decimals for rounding output freq.
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{freq}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:FREQ:SWE:STOP?"),
        ]
        # Act
        frequency = self.instr.get_sweep_stop_frequency()
        # Assert
        self.assertEqual(expected_freq, frequency)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_mode(self):
        """Test setting the sweep_mode."""
        mode = random.randint(0, 3)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:MOD {mode}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_mode(mode)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_mode_excepts(self):
        """Test setting the sweep_mode in with values out-of-bounds."""
        input_mode = [-1, 4]
        # Act
        for mode in input_mode:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_mode(mode)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_mode(self):
        """Test getting the sweep_mode."""
        expected_mode = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_mode}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:MOD?"),
        ]
        # Act
        mode = self.instr.get_sweep_mode()
        # Assert
        self.assertEqual(expected_mode, mode)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_speed(self):
        """Test setting the sweep_speed."""
        speed = random.choice([1, 2, 5, 10, 20, 50, 100, 200])
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:SPE {speed}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_speed(speed)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_speed_excepts(self):
        """Test setting the sweep_speed in with values out-of-bounds."""
        input_speed = [0, 199]
        # Act
        for speed in input_speed:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_speed(speed)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_speed(self):
        """Test getting the sweep_speed in nm/s."""
        expected_speed = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_speed}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:SPE?"),
        ]
        # Act
        speed = self.instr.get_sweep_speed()
        # Assert
        self.assertEqual(expected_speed, speed)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_cycles(self):
        """Test setting the sweep_cycles."""
        cycles = random.randint(0, 999)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:CYCL {cycles}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_cycles(cycles)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_cycles_excepts(self):
        """Test setting the sweep_cycles in with values out-of-bounds."""
        input_cycles = [-1, 1000]
        # Act
        for cycles in input_cycles:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_cycles(cycles)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_cycles(self):
        """Test getting the sweep_cycles."""
        expected_cycles = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_cycles}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:CYCL?"),
        ]
        # Act
        cycles = self.instr.get_sweep_cycles()
        # Assert
        self.assertEqual(expected_cycles, cycles)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_dwell(self):
        """Test setting the sweep_dwell."""
        dwell = float(random.randint(0, 999))
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:DWEL {dwell}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_dwell(dwell)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_dwell_excepts(self):
        """Test setting the sweep_dwell in with values out-of-bounds."""
        input_dwell = [-0.1, 1000.0]
        # Act
        for dwell in input_dwell:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_dwell(dwell)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_dwell(self):
        """Test getting the sweep_dwell."""
        expected_dec = 1  # should be 1 decimal accuracy
        expected_dwell = 2.0
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_dwell:.{expected_dec}f}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:DWEL?"),
        ]
        # Act
        dwell = self.instr.get_sweep_dwell()
        # Assert
        self.assertEqual(expected_dwell, dwell)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_delay(self):
        """Test setting the sweep_delay."""
        delay = float(random.randint(0, 999))
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:DEL {delay}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_delay(delay)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_delay_excepts(self):
        """Test setting the sweep_delay in with values out-of-bounds."""
        input_delay = [-0.1, 1000.0]
        # Act
        for delay in input_delay:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_delay(delay)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_delay(self):
        """Test getting the sweep_delay."""
        expected_dec = 1  # should be 1 decimal accuracy
        expected_delay = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_delay:.{expected_dec}f}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:DEL?"),
        ]
        # Act
        delay = self.instr.get_sweep_delay()
        # Assert
        self.assertEqual(expected_delay, delay)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_state(self):
        """Test setting the sweep_state."""
        state = random.randint(0, 1)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":WAV:SWE:STAT {state}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_sweep_state(state)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_sweep_state_excepts(self):
        """Test setting the sweep_state in with values out-of-bounds."""
        input_state = [-1, 2]
        # Act
        for state in input_state:
            with self.assertRaises(ValueError):
                self.instr.set_sweep_state(state)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_sweep_state(self):
        """Test getting the sweep_state."""
        expected_state = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_state}"
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:STAT?"),
        ]
        # Act
        state = self.instr.get_sweep_state()
        # Assert
        self.assertEqual(expected_state, state)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_timing_mode(self):
        """Test setting the trigger_output_timing_mode."""
        mode = random.randint(0, 3)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":TRIG:OUTP {mode}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_trigger_output_timing_mode(mode)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_timing_mode_excepts(self):
        """Test setting the trigger_output_timing_mode in with values out-of-bounds."""
        input_mode = [-1, 4]
        # Act
        for mode in input_mode:
            with self.assertRaises(ValueError):
                self.instr.set_trigger_output_timing_mode(mode)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_trigger_output_timing_mode(self):
        """Test getting the trigger_output_timing_mode."""
        expected_mode = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_mode}"
        ]
        expected_ask_calls = [
            call().ask(":TRIG:OUTP?"),
        ]
        # Act
        mode = self.instr.get_trigger_output_timing_mode()
        # Assert
        self.assertEqual(expected_mode, mode)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_period_mode(self):
        """Test setting the trigger_output_period_mode."""
        mode = random.randint(0, 1)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":TRIG:OUTP:SETT {mode}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_trigger_output_period_mode(mode)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_period_mode_excepts(self):
        """Test setting the trigger_output_period_mode in with values out-of-bounds."""
        input_mode = [-1, 2]
        # Act
        for mode in input_mode:
            with self.assertRaises(ValueError):
                self.instr.set_trigger_output_period_mode(mode)

        # Assert
        self._scpi_mock.assert_not_called()

    def test_get_trigger_output_period_mode(self):
        """Test getting the trigger_output_period_mode."""
        expected_mode = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_mode}"
        ]
        expected_ask_calls = [
            call().ask(":TRIG:OUTP:SETT?"),
        ]
        # Act
        mode = self.instr.get_trigger_output_period_mode()
        # Assert
        self.assertEqual(expected_mode, mode)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_step(self):
        """Test setting the trigger_output_step."""
        dec = 4  # Response in 4 decimal accuracy
        sweep_stop_wl, sweep_start_wl = 1360.0, 1280.0
        wl_range = sweep_stop_wl - sweep_start_wl
        step = random.uniform(0.0001, wl_range)
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{sweep_stop_wl}",
            f"{sweep_start_wl}",
            '0,"No error"',
            "No alerts."
        ]
        expected_write_calls = [
            call().write(f":TRIG:OUTP:STEP {step:.{dec}f}")
        ]
        expected_ask_calls = [
            call().ask(":SYST:ERR?"),
            call().ask(":SYST:ALER?")
        ]
        # Act
        self.instr.set_trigger_output_step(step)
        # Assert
        self._scpi_mock.assert_has_calls(expected_write_calls)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_set_trigger_output_step_excepts(self):
        """Test setting the trigger_output_step in with values out-of-bounds."""
        sweep_stop_wl, sweep_start_wl = 1360.0, 1280.0
        wl_range = sweep_stop_wl - sweep_start_wl
        input_step = [0.0000, wl_range + 0.0001]
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{sweep_stop_wl}",
            f"{sweep_start_wl}",
            f"{sweep_stop_wl}",
            f"{sweep_start_wl}",
        ]
        expected_ask_calls = [
            call().ask(":WAV:SWE:STOP?"),
            call().ask(":WAV:SWE:STAR?"),
            call().ask(":WAV:SWE:STOP?"),
            call().ask(":WAV:SWE:STAR?")
        ]
        # Act
        for step in input_step:
            with self.assertRaises(ValueError):
                self.instr.set_trigger_output_step(step)

        # Assert
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_get_trigger_output_step(self):
        """Test getting the trigger_output_step."""
        expected_dec = 1  # should be 1 decimal accuracy
        expected_step = 2
        self._scpi_mock(self._transport_mock).ask.side_effect = [
            f"{expected_step:.{expected_dec}f}"
        ]
        expected_ask_calls = [
            call().ask(":TRIG:OUTP:STEP?"),
        ]
        # Act
        step = self.instr.get_trigger_output_step()
        # Assert
        self.assertEqual(expected_step, step)
        self._scpi_mock.assert_has_calls(expected_ask_calls)

    def test_issue_soft_trigger(self):
        """Test issue_soft_trigger."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ['0,"No error"', "No alerts."]

        self.instr.issue_soft_trigger()

        self._scpi_mock.assert_has_calls([call().write(":TRIG:INP:SOFT")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ERR?")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ALER?")])

    def test_readout_points(self):
        """Test getting number of readout points."""
        readout_points = 124
        self._scpi_mock(self._transport_mock).ask.side_effect = [f"{readout_points}"]

        points = self.instr.readout_points()

        self.assertEqual(readout_points, points)
        self._scpi_mock.assert_has_calls([call().ask(":READ:POIN?")])

    def test_readout_data(self):
        """Test reading out data."""
        # Let's make 200 data point readings
        expected_data_points = 200
        expected_data = [round(random.uniform(self.wl_min, self.wl_max), 4) for _ in range(expected_data_points)]
        binary_data = b''.join([int(d * 1e4).to_bytes(4, byteorder="little") for d in expected_data])
        self._scpi_mock(self._transport_mock).read_binary_data.side_effect = [
            binary_data
        ]
        expected_write_calls = [
            call().write(":READout:DATa?"),
        ]
        # Act
        data = self.instr.readout_data()
        # Assert
        # Use numpy list assertion as otherwise some float inaccuracy errors appear
        np.testing.assert_allclose(expected_data, data, rtol=1e-4)
        self._scpi_mock.assert_has_calls(expected_write_calls)

    def test_shutdown(self):
        """Test shutdown."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ['0,"No error"', "No alerts."]

        self.instr.shutdown()

        self._scpi_mock.assert_has_calls([call().write(":SPEC:SHUT")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ERR?")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ALER?")])

    def test_reboot(self):
        """Test reboot."""
        self._scpi_mock(self._transport_mock).ask.side_effect = ['0,"No error"', "No alerts."]

        self.instr.reboot()

        self._scpi_mock.assert_has_calls([call().write(":SPECial:REBoot")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ERR?")])
        self._scpi_mock.assert_has_calls([call().ask(":SYST:ALER?")])



"""
Stanford Research Systems; voltage source, SIM system mainframe and diode temperature monitor.

The qmi.instruments.picotech package provides support for:
- DC205 voltage source
- SIM900 mainframe
- SIM922 diode temperature monitor.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.stanford_research_systems.dc205 import SRS_DC205 as Srs_Dc205
from qmi.instruments.stanford_research_systems.sim900 import Sim900 as Srs_Sim900
from qmi.instruments.stanford_research_systems.sim922 import SIM922 as Srs_Sim922



"""Driver for the adbasic command line script.

It is recommended to use this script instead of directly invoking /opt/adwin/bin/adbasic.
It provides much nicer formatting of compiler errors, a working exit code, and help.
"""

import argparse
import logging
import os
from pathlib import Path, WindowsPath
import re
import sys
import shutil
import subprocess
import time
from typing import List, NamedTuple, Optional, Tuple

import colorama

from qmi.core.exceptions import QMI_RuntimeException


# Value used to mark high process priority.
PRIO_HIGH = 1000

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

LINUX_INCLUDE_DIR = "share/adbasic/Inc"
LINUX_LIBRARY_DIR = "share/adbasic/Lib"
# Determine default ADwin software location based on OS.
if "linux" in sys.platform or "darwin" in sys.platform:
    DEFAULT_ADWINDIR = "/opt/adwin"
    ADBASIC_COMPILER_EXECUTABLE = {"T11": "bin/adbasic", "T12": "bin/adbasic", "T12.1": "bin/adbasic"}

else:
    DEFAULT_ADWINDIR = "C:\\ADwin"
    ADBASIC_COMPILER_EXECUTABLE = {
        "T11": "ADbasic\\ADbasicCompiler.exe",
        "T12": "ADbasic\\ADbasic_C.exe",
        "T12.1": "ADbasic\\ADbasic_C.exe"
    }


class AdbasicError(NamedTuple):
    """Information from one error message from the ADbasic compiler."""
    error_number: int
    error_description: str
    error_line: str
    filename: str
    line_number: int

    def __str__(self) -> str:
        msg = ""
        if self.filename:
            msg += "[" + self.filename
            if self.line_number:
                msg += ":{}".format(self.line_number)
            msg += "] "
        if self.error_number:
            msg += "Error {}: ".format(self.error_number)
        msg += self.error_description
        if self.error_line:
            msg += " ({!r})".format(self.error_line.strip())
        return msg


class AdbasicResult(NamedTuple):
    """Information from a run of the ADbasic compiler."""
    command_line: str
    duration: float
    returncode: int
    success: bool
    errors: List[AdbasicError]
    warnings: List[AdbasicError]


class AdbasicCompilerException(QMI_RuntimeException):
    """Raised when an error occurs while compiling ADbasic code."""

    def __init__(self, message: str, errors: List[AdbasicError]) -> None:
        super().__init__([message, errors])
        self.adbasic_errors = errors
        self.message = message

    def __str__(self) -> str:
        error_message = self.message
        for error in self.adbasic_errors:
            error_message += "\n  " + str(error)
        return error_message
    
    def __reduce__(self):
        return (AdbasicCompilerException, (self.message, self.adbasic_errors))


def _extract_lines(lines_bin: bytes) -> List[str]:
    """Extract lines from compiler output stream."""
    lines = lines_bin.decode("ascii", errors='replace').split("\r\n")
    while len(lines) > 0 and lines[-1] in ["", "\n"]:
        lines = lines[:-1]
    return lines


def _parse_stderr_lines(stderr_lines: List[str]) -> Tuple[List[AdbasicError], List[AdbasicError]]:
    """Parse stderr output from the compiler."""

    error_line_v1_pattern_string = r"Error: No: ([0-9]+) (.*?) line: (.*?) file: (.*?) line no\.: ([0-9]+) "
    error_line_v1_pattern = re.compile(error_line_v1_pattern_string)

    error_line_v2_pattern_string = r"Error: No: ([0-9]+) (.*?) file: ([^ ]*) "
    error_line_v2_pattern = re.compile(error_line_v2_pattern_string)

    error_invalid_option_pattern_string = r"Invalid Option: ([^ ]+)"
    error_invalid_option_pattern        = re.compile(error_invalid_option_pattern_string)

    compilation_aborted_line_pattern_string = r"Compilation aborted !"
    compilation_aborted_line_pattern = re.compile(compilation_aborted_line_pattern_string)

    error_warning_count_line_pattern_string = r"([0-9]+) error.s., ([0-9]+) warning.s."
    error_warning_count_line_pattern = re.compile(error_warning_count_line_pattern_string)

    fixme_line_pattern_string = r"([0-9a-fA-F]+):fixme:([:a-zA-Z]+) .+"
    fixme_warning_count_line_pattern = re.compile(fixme_line_pattern_string, re.DOTALL)

    errors: List[AdbasicError] = []
    warnings: List[AdbasicError] = []

    for line in stderr_lines:

        # Parse stderr line(s).
        while len(line) != 0:
            match = error_line_v1_pattern.match(line)
            if match:
                (error_number, error_description, error_line, filename, line_number) = match.groups()
                error_number_int = int(error_number)
                line_number_int = int(line_number)
                error = AdbasicError(error_number_int, error_description, error_line, filename, line_number_int)
                errors.append(error)
                line = line[match.end():]
                continue

            match = error_line_v2_pattern.match(line)
            if match:
                (error_number, error_description, filename) = match.groups()
                error_number_int = int(error_number)
                error = AdbasicError(error_number_int, error_description, "", filename, 0)
                errors.append(error)
                line = line[match.end():]
                continue

            match = error_invalid_option_pattern.match(line)
            if match:
                (invalid_option, ) = match.groups()
                error = AdbasicError(0, "Invalid command line option: {}".format(invalid_option), "", "", 0)
                errors.append(error)
                line = line[match.end():]
                continue

            match = compilation_aborted_line_pattern.match(line)
            if match:
                error_description = match.group(0)
                error = AdbasicError(0, error_description, "", "", 0)
                errors.append(error)
                line = line[match.end():]
                continue

            match = error_warning_count_line_pattern.match(line)
            if match:
                error_count = int(match.group(1))
                warning_count = int(match.group(2))
                if (error_count > 0) or (warning_count > 0):
                    error = AdbasicError(0, match.group(0), "", "", 0)
                    warnings.append(error)
                line = line[match.end():]
                continue

            match = fixme_warning_count_line_pattern.match(line)
            if match:
                warning_count = int("0x" + match.group(1), 16)
                if warning_count > 0:
                    error = AdbasicError(0, match.group(0), "", "", 0)
                    warnings.append(error)
                line = line[match.end():]
                continue

            raise AdbasicCompilerException("Unrecognized ADbasic error string: {!r}".format(line), [])

    return errors, warnings


def _library_path_arguments(adwin_dir: str) -> List[str]:
    """Return ADbasic command-line arguments to specify the include path and library path."""

    # The Linux version of the compiler needs an explicit path to the standard library.
    if "linux" in sys.platform or "darwin" in sys.platform:
        include_file_path = Path(adwin_dir, LINUX_INCLUDE_DIR)
        library_file_path = Path(adwin_dir, LINUX_LIBRARY_DIR)
        include_path_argument = "/IP" + include_file_path.as_posix()
        library_path_argument = "/LP" + library_file_path.as_posix()
        return [include_path_argument, library_path_argument]
    else:
        return []


def run_adbasic_compiler(adwin_dir: str,
                         processor_type: str,
                         adbasic_arguments: List[str],
                         working_dir: Optional[str] = None,
                         parse_stderr: bool = True,
                         remove_c_directory: bool = False,
                         pretty_print: bool = False
                         ) -> AdbasicResult:
    """Run the ADbasic compiler.

    Parameters:
        adwin_dir:      Base directory of ADwin software installation (e.g. /opt/adwin).
        processor_type: The type of processor the command is run for.
        adbasic_arguments: List of command-line options for the ADbasic compiler.
        working_dir:    Working directory when running the compiler.
        parse_stderr:   True to parse error messages from the ADbasic compiler and return them.
                        False to ignore error messages from the compiler.
        remove_c_directory: True to remove the directory with intermediate C files from the compiler.
        pretty_print:   True to print compiler messages in pretty colors.
                        False to log messages via the Python log system.

    Returns:
        AdbasicResult tuple containing results from the compilation process.
        If `parse_stderr` is True, this result will contain parsed compiler error messages.

    Raises:
        AdbasicCompilerException: If an unexpected error occurs while interacting with the ADbasic compiler.
    """
    # Determine compiler command
    compiler_command = ADBASIC_COMPILER_EXECUTABLE[processor_type]
    # Determine path to ADbasic compiler.
    adbasic_executable_p = Path(adwin_dir, compiler_command)
    if "linux" in sys.platform or "darwin" in sys.platform:
        adbasic_executable = adbasic_executable_p.as_posix()

    else:
        adbasic_executable = str(adbasic_executable_p).replace("/", "\\")

    run_args = [adbasic_executable] + adbasic_arguments
    _logger.debug("Running %s", " ".join("{!r}".format(arg) for arg in run_args))

    command_line = " ".join(run_args)

    t1 = time.monotonic()

    # pylint: disable=subprocess-run-check
    completed_process = subprocess.run(args=run_args,
                                       capture_output=True,
                                       cwd=working_dir)

    t2 = time.monotonic()
    duration = (t2 - t1)

    if remove_c_directory:
        if working_dir:
            temp_c_dir = os.path.join(working_dir, "c")
        else:
            temp_c_dir = "c"
        try:
            shutil.rmtree(temp_c_dir)
        except FileNotFoundError:
            pass

    stdout_lines = _extract_lines(completed_process.stdout)
    stderr_lines = _extract_lines(completed_process.stderr)

    for line in stdout_lines:
        if pretty_print:
            print(colorama.Fore.GREEN + "(stdout)" + colorama.Style.RESET_ALL, line)
        else:
            _logger.debug("ADbasic stdout: %s", line)

    if parse_stderr:

        # Parse errors and warnings.
        (errors, warnings) = _parse_stderr_lines(stderr_lines)

        # Log parsed errors.
        for error in errors:
            if pretty_print:
                print("{}(error){}".format(colorama.Fore.RED, colorama.Style.RESET_ALL), str(error))
            else:
                _logger.error("ADbasic error: %s", str(error))
        for error in warnings:
            if pretty_print:
                print("{}(warning){}".format(colorama.Fore.RED, colorama.Style.RESET_ALL), str(error))
            else:
                _logger.warning("ADbasic warning: %s", str(error))

    else:
        errors = []
        warnings = []

        # Log raw errors.
        for line in stderr_lines:
            if pretty_print:
                print(colorama.Fore.RED + "(stderr)" + colorama.Style.RESET_ALL, line)
            else:
                _logger.error("ADbasic stderr: %s", line)

    # The returncode may be 0 even if compilation failed, especially under Linux.
    # We therefore assume the compiler was successful only if returncode is 0
    # and there are no messages on the stderr stream.
    success = (completed_process.returncode == 0) and (not errors)

    return AdbasicResult(command_line, duration, completed_process.returncode, success, errors, warnings)


def print_adbasic_compiler_help(adwin_dir: str, processor_type: str) -> AdbasicResult:
    # The /<processor_type> is needed to select the ADbasic-to-C compile path.
    adbasic_arguments = [f"/P{processor_type.lstrip('T').replace('.', '')}", "/H"]
    return run_adbasic_compiler(adwin_dir,
                                processor_type,
                                adbasic_arguments,
                                working_dir=None,
                                parse_stderr=False,
                                remove_c_directory=False,
                                pretty_print=True)


def compile_program(basic_filename: str,
                    trigger: str,
                    process_number: int,
                    processor_type: str,
                    hardware_type: str,
                    priority: int = PRIO_HIGH,
                    working_dir: Optional[str] = None,
                    keep_c_files: bool = False,
                    adwin_dir: str = DEFAULT_ADWINDIR,
                    pretty_print: bool = False
                    ) -> AdbasicResult:
    """Compile an ADbasic program.

    The ADbasic compiler expects that include files can be found relative
    to the current working directory.

    If compilation is successful, an executable file will be created in
    the same directory as the source file. The name of this file is based
    on the name of the source file, with the ".bas" extension replaced
    by ".TCn" depending on the process number.

    Parameters:
        basic_filename: File name of the ADbasic .bas file to compile.
        trigger:        Select event trigger mechanism, either "external" or "timer".
        process_number: Select process number, range 1 to 10.
        processor_type: Select processor type "T<x>", where <x> is 11, 12 or 12.1.
        hardware_type:  The hardware that is used for compilation. Possible choices P, PII, G, GII (P=pro, G=Gold)
        priority:       Process priority, integer in range -10 to 10 for low priority,
                        or PRIO_HIGH for high priority.
        working_dir:    Working directory when running the compiler.
                        This determines the base directory for searching include files.
                        Pass None to keep the current working directory of the process.
        keep_c_files:   Keep 'c' directory containing intermediate C files from ADbasic compiler.
                        This only works under Linux.
        adwin_dir:      Base directory of ADwin software installation.
        pretty_print:   True to print compiler messages in pretty colors.
                        False to log messages via the Python log system.

    Returns:
        AdbasicResult tuple containing results and error messages (if any) from the compilation process.
        If compilation is successful, the `success` field of this tuple will be True.
        If compilation fails, the `success` field will be False and the `errors` field will contain
        a list of error messages from the compiler.

    Raises:
        AdbasicCompilerException: If an unexpected error occurs while interacting with the ADbasic compiler.
    """
    # The /M instructs the compiler to produce an executable.
    # The /<processor_type> is needed to select the ADbasic-to-C compile path.

    if trigger == "external":
        trigger_argument = "/EE"
    else:
        trigger_argument = "/ET"

    if priority == PRIO_HIGH:
        priority_argument = "/PH"
    elif -10 <= priority <= 10:
        priority_argument = "/PL{}".format(int(priority))
    else:
        raise ValueError("Invalid process priority {!r}".format(priority))

    process_number_argument = "/PN{}".format(process_number)
    optimization_level_argument = "/O2"

    adbasic_arguments = ["/M",
                         f"/P{processor_type.lstrip('T').replace('.', '')}",
                         f"/S{hardware_type}",  # Adwin type, SPII for Pro II, SGII for Gold II
                         trigger_argument,
                         process_number_argument,
                         priority_argument,
                         optimization_level_argument]
    adbasic_arguments += _library_path_arguments(adwin_dir)
    adbasic_arguments.append(basic_filename)

    return run_adbasic_compiler(adwin_dir,
                                processor_type,
                                adbasic_arguments,
                                working_dir,
                                parse_stderr=True,
                                remove_c_directory=not keep_c_files,
                                pretty_print=pretty_print)


def compile_library(basic_filename: str,
                    processor_type: str,
                    hardware_type: str,
                    working_dir: Optional[str] = None,
                    keep_c_files: bool = False,
                    adwin_dir: str = DEFAULT_ADWINDIR,
                    pretty_print: bool = False
                    ) -> AdbasicResult:

    # The /L instructs the compiler to produce a library.
    # The /<processor_type> is needed to select the ADbasic-to-C compile path.
    # The /<hardware_type> is needed for the compiler info.
    # The optimization level argument is "/O2"

    adbasic_arguments = [
        "/L",
        f"/P{processor_type.lstrip('T').replace('.', '')}",
        f"/S{hardware_type}",
        "/O2"
    ]
    adbasic_arguments += _library_path_arguments(adwin_dir)
    adbasic_arguments.append(basic_filename)

    return run_adbasic_compiler(adwin_dir,
                                processor_type,
                                adbasic_arguments,
                                working_dir,
                                parse_stderr=True,
                                remove_c_directory=not keep_c_files,
                                pretty_print=pretty_print)


def check_compiler_result(basic_filename: str, result: AdbasicResult) -> None:
    """Check that the compiler run was successful.

    Parameters:
        basic_filename: File name of the ADbasic .bas file.
        result:         Result from running the compiler.

    Raises:
        AdbasicCompilerException: If the compiler was not successful.
    """
    if not result.success:
        raise AdbasicCompilerException("ADbasic compiler reported error when compiling {!r} (exit status {})"
                                       .format(basic_filename, result.returncode),
                                       result.errors)


def main():

    exitcode = 0

    DEFAULT_TRIGGER_MECHANISM = "external"
    DEFAULT_PROCESSOR_TYPE = "T12"
    DEFAULT_HARDWARE_TYPE = "PII"
    DEFAULT_PROCESS_NUMBER = 1

    parser = argparse.ArgumentParser(
        description="Python wrapper for the ADbasic script, with improved error reporting.")

    parser.add_argument("--adwin-dir", default=DEFAULT_ADWINDIR,
                        help="base directory of ADwin software installation (default: {!r})".format(DEFAULT_ADWINDIR))
    parser.add_argument("--compiler-help", action="store_true",
                        help="show help of the underlying ADbasic compiler")
    parser.add_argument("-k", "--keep-c-files", action="store_true",
                        help="keep local 'c' directory containing intermediate C files from ADbasic compiler")
    parser.add_argument("-l", "--library", action="store_true",
                        help="compile a library instead of a binary executable")
    parser.add_argument("-P", "--processor-type", choices=["T11", "T12", "T12.1"],
                        default=DEFAULT_PROCESSOR_TYPE,
                        help="select ADwin processor type (default: {!r})".format(DEFAULT_PROCESSOR_TYPE))
    parser.add_argument("-H", "--hardware-type", choices=["P", "PII", "G", "GII"],
                        default=DEFAULT_HARDWARE_TYPE,
                        help="select ADwin hardware type (default: {!r})".format(DEFAULT_HARDWARE_TYPE))
    parser.add_argument("-t", "--trigger", choices=["external", "timer"], default=DEFAULT_TRIGGER_MECHANISM,
                        help="select event trigger mechanism (default: {!r})".format(DEFAULT_TRIGGER_MECHANISM))
    parser.add_argument("-p", "--process-number", type=int, default=DEFAULT_PROCESS_NUMBER,
                        help="select process number, range 1..10 (default: {})".format(DEFAULT_PROCESS_NUMBER))
    parser.add_argument("filename", nargs='?', help="file name of the adbasic .bas file")

    args = parser.parse_args()

    colorama.init()
    try:
        result = None
        if args.compiler_help:
            result = print_adbasic_compiler_help(args.adwin_dir, args.processor_type)
        elif args.filename is None:
            parser.print_usage()
        elif args.library:
            result = compile_library(args.filename,
                                     processor_type=args.processor_type,
                                     hardware_type=args.hardware_type,
                                     working_dir=None,
                                     keep_c_files=args.keep_c_files,
                                     adwin_dir=args.adwin_dir,
                                     pretty_print=True)
            if not result.success:
                exitcode = 1
        else:
            result = compile_program(args.filename,
                                     args.trigger,
                                     args.process_number,
                                     args.processor_type,
                                     hardware_type=args.hardware_type,
                                     priority=PRIO_HIGH,
                                     working_dir=None,
                                     keep_c_files=args.keep_c_files,
                                     adwin_dir=args.adwin_dir,
                                     pretty_print=True)
            if not result.success:
                exitcode = 1

        if result is not None:
            print()
            print("Executed command line:")
            print()
            print("    {}".format(result.command_line))
            print()
            print("Execution time: {:.3f} seconds.".format(result.duration))
            print()
            print("Performing exit({}), bye!".format(exitcode))
            print()
    finally:
        colorama.deinit()

    return exitcode


if __name__ == "__main__":
    sys.exit(main())



"""
Wieserlabs Electronics, RF source.

The qmi.instruments.wieserlabs package provides support for:
- WL-FlexDDS-NG.
"""
from qmi.instruments.wieserlabs.flexdds import OutputChannel, DdsRegister, DcpRegister, PllStatus
# Alternative, QMI naming convention approved names
from qmi.instruments.wieserlabs.flexdds import Wieserlabs_FlexDDS_NG_Dual as Wieserlabs_FlexDdsNg



"""Instrument driver for the Newport Agilis AG-UC8 Piezo Stepper Controller."""

import enum
import logging
import re
import time
from typing import Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

# Global variable holding the logger for this module.
from qmi.core.transport import create_transport

_logger = logging.getLogger(__name__)


class AxisStatus(enum.IntEnum):
    """Axis status codes."""

    READY = 0
    STEPPING = 1
    JOGGING = 2
    MOVING_TO_LIMIT = 3


class Newport_AG_UC8(QMI_Instrument):
    """Instrument driver for the Newport AG-UC8 Piezo Stepper Controller."""

    _rpc_constants = ["ERROR_CODES", "SPEED_TABLE"]

    # Meaning of error codes returned by the device.
    ERROR_CODES = {
        0: "No error",
        -1: "Unknown command",
        -2: "Axis out of range",
        -3: "Wrong format for parameter",
        -4: "Parameter out of range",
        -5: "Not allowed in local mode",
        -6: "Not allowed in current state",
    }

    # Meaning of speed settings.
    SPEED_TABLE = {
        1: "5 steps/second at defined step amplitude",
        2: "100 steps/second at maximum step amplitude",
        3: "1700 steps/second at maximum step amplitude",
        4: "666 steps/second at defined step amplitude",
    }

    # By default, expect response to command within 1 second.
    RESPONSE_TIMEOUT = 1.0

    # Some commands (position measurement and absolute move) can take
    # up to 2 minutes to complete.
    SLOW_RESPONSE_TIMEOUT = 120.0

    # After a command which does not generate a response, a short
    # delay is needed before we can send the following command.
    COMMAND_DELAY = 0.005

    # Delay after changing channels.
    CHANNEL_SWITCH_DELAY = 0.02

    # After a reset command, a longer delay is needed before
    # we can send the following command.
    RESET_DELAY = 0.05

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver.

        The instrument is typically accessed via the USB serial port,
        e.g. "serial:/dev/ttyUSBn" or "serial:COM3".

        :argument name: Name for this instrument instance.
        :argument transport: QMI transport descriptor for command interface.
        """
        super().__init__(context, name)
        self._transport = create_transport(
            transport,
            default_attributes={
                "baudrate": 921600,
                "bytesize": 8,
                "parity": "N",
                "stopbits": 1,
            },
        )
        self._current_channel: Optional[int] = None

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

        # Do not assume any specific channel is selected.
        self._current_channel = None
        # Set controller in remote mode (otherwise many commands don't work).
        self._write("MR")
        # Check error status.
        self._check_error("MR")

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        self._transport.close()
        super().close()

    def _write(self, cmd: str) -> None:
        """Send command to instrument and check instrument responds with "OK"."""
        self._check_is_open()

        # Send command, terminated with \r\n.
        rawcmd = cmd.encode("ascii") + b"\r\n"
        self._transport.write(rawcmd)

        # Sleep until command completed.
        time.sleep(self.COMMAND_DELAY)

    def _ask(self, cmd: str, timeout: float) -> str:
        """Send command to instrument and return response from instrument."""
        self._check_is_open()

        # Send command, terminated with \r\n.
        rawcmd = cmd.encode("ascii") + b"\r\n"
        self._transport.write(rawcmd)

        # Read response.
        resp = self._transport.read_until(message_terminator=b"\n", timeout=timeout)
        resp = resp.rstrip(b"\r\n")

        return resp.decode("ascii", errors="replace")

    def _get_attribute(self, cmd: str, timeout: float = RESPONSE_TIMEOUT) -> int:
        """Send the specified command to the device and expect an integer value as result.

        This function expects that the device answers with a single line,
        containing an echo of the command immediately followed by an integer value.

        :argument cmd: Command to send to the device.
        :return: Integer value returned by the device.
        :raises QMI_InstrumentException: If the device sends an unexpected answer.
        """
        resp = self._ask(cmd, timeout=timeout)
        basecmd = re.sub(r"[0-9?]+$", "", cmd, count=1)  # strip argument and '?' marker
        if resp.startswith(basecmd):
            try:
                return int(resp[len(basecmd):].strip())
            except ValueError:
                # Invalid numeric format.
                # Ignore error; it will be reported below.
                pass

        _logger.warning("Unexpected response to command %s: %r", cmd, resp)
        raise QMI_InstrumentException(f"Unexpected response to command {cmd}: '{resp}'")

    @rpc_method
    def get_last_error(self) -> int:
        """Send a TE command (get error of previous command) and return
        a numerical error code. This function is called automatically after each command sent
        to the device.

        Raises:
            QMI_InstrumentException: When a command results in error.

        Returns:
            error: Integer error code for previous command. Value 0 means no error (success).
                   See class attribute ERROR_CODES for the meaning of the codes.
        """
        return self._get_attribute("TE")

    def _check_error(self, cmd: str) -> None:
        """Check the error status of the motor controller.

        :argument cmd: Most recent command.
        :raises QMI_InstrumentException: If the controller returns an error code.
        """
        err = self.get_last_error()
        if err != 0:
            errstr = self.ERROR_CODES.get(err, "unknown")
            _logger.warning("Command %s failed with error %d (%s)", cmd, err, errstr)
            raise QMI_InstrumentException(
                f"Command {cmd} failed with error {err} ({errstr})"
            )

    def _get_axis_attribute(
        self, attr: str, axis: int, timeout: float = RESPONSE_TIMEOUT
    ) -> int:
        """Read the specified axis attribute and return an integer value.

        Parameters:
            attr: The axis attribute to be obtained.
            axis: The axis number. Must be 1 or 2.
            timeout: Timeout for instrument response.

        Raises:
            QMI_UsageException: If invalid axis index is given as input.

        Returns:
            The value of the queried attribute on selected axis.
        """
        if axis not in (1, 2):
            raise QMI_UsageException(f"Invalid axis index {axis}")
        return self._get_attribute(f"{axis}{attr}", timeout)

    def _set_axis_attribute(self, attr: str, axis: int, value: int) -> None:
        """Set the specified axis attribute.

        Parameters:
            attr: The axis attribute to be set.
            axis: The axis number. Must be 1 or 2.
            value: An integer value for the attribute.

        Raises:
            QMI_UsageException: If invalid axis index is given as input.
        """
        if axis not in (1, 2):
            raise QMI_UsageException(f"Invalid axis index {axis}")

        cmd = f"{axis}{attr}{value}"
        self._write(cmd)
        self._check_error(cmd)

    @rpc_method
    def reset(self) -> None:
        """Reset the motor controller. Re-enables remote mode afterwards"""
        _logger.info("[%s] reset", self._name)
        self._current_channel = None
        self._write("RS")
        # Sleep until reset completed.
        time.sleep(self.RESET_DELAY)
        # Switch controller to remote mode (many commands require remote mode).
        self._write("MR")
        # Check error status.
        self._check_error("MR")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._ask("VE", timeout=self.RESPONSE_TIMEOUT)
        words = resp.strip().split()
        if len(words) == 2:
            model = words[0]
            version = words[1]
        else:
            _logger.warning("Unexpected response to VE command: %r", resp)
            raise QMI_InstrumentException(
                f"Unexpected response to VE command: '{resp}'"
            )
        return QMI_InstrumentIdentification(
            vendor="Newport", model=model, serial=None, version=version
        )

    @rpc_method
    def select_channel(self, channel: int) -> None:
        """Select the specified channel.

        The selected channel is attached to the motor drive electronics.
        This function is called automatically by methods which have a
        "channel" parameter.

        Changing the channel is only allowed when the axis is not moving.

        Parameters:
            channel: The channel number. Must in range 1-4.

        Raises:
            QMI_UsageException: By invalid channel number input.
        """

        # Check channel index.
        if channel not in (1, 2, 3, 4):
            raise QMI_UsageException(f"Invalid channel index {channel}")

        # Select channel, unless the channel is already selected.
        if self._current_channel != channel:
            self._write(f"CC{channel}")
            self._check_error("CC")
            time.sleep(self.CHANNEL_SWITCH_DELAY)
            self._current_channel = channel

    @rpc_method
    def get_limit_status(self, channel: int) -> Tuple[bool, bool]:
        """Get the limit switch status for the specified channel.

        :argument channel: Channel index (range 1 .. 4).
        :raises: QMI_InstrumentException if invalid response is received.
        :return: Tuple (axis1_limit_active, axis2_limit_active).
        """
        self.select_channel(channel)
        value = self._get_attribute("PH")
        if value == 0:
            return (False, False)
        elif value == 1:
            return (True, False)
        elif value == 2:
            return (False, True)
        elif value == 3:
            return (True, True)
        raise QMI_InstrumentException(f"Invalid response to PH command: '{value}'")

    @rpc_method
    def get_step_delay(self, axis: int) -> int:
        """Get step delay.
        :argument axis: Axis index (1 .. 2).
        :return: Step delay in units of 10 us.
        """
        return self._get_axis_attribute("DL?", axis)

    @rpc_method
    def set_step_delay(self, axis: int, value: int) -> None:
        """Set step delay.

        :argument axis: Axis index (1 .. 2).
        :argument value: Step delay in units of 10 us (range 1 .. 200000).
        """
        self._set_axis_attribute("DL", axis, value)

    @rpc_method
    def get_step_amplitude(self, axis: int, direction: int) -> int:
        """Get step amplitude.

        :argument axis: Axis index (1 .. 2).
        :argument direction: Direction (0=positive, 1=negative).
        :raises: QMI_UsageException by invalid direction index number.
        :return: Step amplitude (1 .. 50).
        """
        if direction == 0:
            attr = "SU+?"
        elif direction == 1:
            attr = "SU-?"
        else:
            raise QMI_UsageException(f"Invalid direction index {direction}")
        return self._get_axis_attribute(attr, axis)

    @rpc_method
    def set_step_amplitude(self, axis: int, direction: int, value: int) -> None:
        """Set step amplitude.

        :argument axis: Axis index (1 .. 2).
        :argument direction: Step direction (0=positive, 1=negative).
        :argument value: New step amplitude (1 .. 50).
        :raises: QMI_UsageException by invalid direction index number.
        """
        if direction == 0:
            attr = "SU+"
        elif direction == 1:
            attr = "SU-"
        else:
            raise QMI_UsageException(f"Invalid direction index {direction}")
        self._set_axis_attribute(attr, axis, value)

    @rpc_method
    def get_step_count(self, axis: int) -> int:
        """Get accumulated number of steps since last reset of the step counter.

        :argument axis: Axis index (1 .. 2).
        :return: Number of steps (difference between positive and negative steps).
        """
        return self._get_axis_attribute("TP", axis)

    @rpc_method
    def clear_step_count(self, axis: int) -> None:
        """Reset the step counter to zero.

        :argument axis: Axis index (1 .. 2).
        :raises: QMI_UsageException by invalid axis index number.
        """
        if axis not in (1, 2):
            raise QMI_UsageException(f"Invalid axis index {axis}")
        cmd = f"{axis}ZP"
        self._write(cmd)
        self._check_error(cmd)

    @rpc_method
    def get_axis_status(self, axis: int) -> AxisStatus:
        """Get status of the axis.

        :argument axis: Axis index (1 .. 2).
        :return: Axis status.
        """
        value = self._get_axis_attribute("TS", axis)
        return AxisStatus(value)

    @rpc_method
    def jog(self, channel: int, axis: int, speed: int) -> None:
        """Start moving in specified direction at specified speed.

        :argument channel: Channel index (1 .. 4).
        :argument axis: Axis index (1 .. 2).
        :argument speed: Speed and direction.
            0 = stop moving.
            positive values 1 .. 4 = move in positive direction.
            negative values -1 .. -4 = move in negative direction.
            See Newport_AG_UC8.SPEED_TABLE for the speed levels.
        """
        self.select_channel(channel)
        self._set_axis_attribute("JA", axis, speed)

    @rpc_method
    def move_limit(self, channel: int, axis: int, speed: int) -> None:
        """Start moving to positive or negative limit.

        :argument channel: Channel index (1 .. 4).
        :argument axis: Axis index (1 .. 2).
        :argument speed: Speed and direction.
            0 = stop moving.
            positive values 1 .. 4 = move in positive direction.
            negative values -1 .. -4 = move in negative direction.
            See Newport_AG_UC8.SPEED_TABLE for the speed levels.
        """
        self.select_channel(channel)
        self._set_axis_attribute("MV", axis, speed)

    @rpc_method
    def measure_position(self, channel: int, axis: int) -> int:
        """Measure current position along the specified axis.

        This is a slow command which may take up to 2 minutes to finish.
        It moves the axis to both limits, then back to its (approximate)
        original position.

        :argument channel: Channel index (1 .. 4).
        :argument axis: Axis index (1 .. 2).
        :return: Position in range 0 .. 1000 representing the current
            position in units of 1/1000 of the total axis range.
        """
        self.select_channel(channel)
        return self._get_axis_attribute("MA", axis, timeout=self.SLOW_RESPONSE_TIMEOUT)

    @rpc_method
    def move_abs(self, channel: int, axis: int, position: int) -> int:
        """Move to absolute position.

        This is a slow command which may take up to 2 minutes to finish.
        It moves the axis to both limits, then to its destination position.

        :argument channel: Channel index (1 .. 4).
        :argument axis: Axis index (1 .. 2).
        :argument position: Target position in range 0 .. 1000, in units
            of 1/1000 of the total axis range.
        """
        self.select_channel(channel)
        return self._get_axis_attribute(
            f"PA{position}", axis, timeout=self.SLOW_RESPONSE_TIMEOUT
        )

    @rpc_method
    def move_rel(self, channel: int, axis: int, steps: int) -> None:
        """Start relative movement.

        :argument channel: Channel index (1 .. 4).
        :argument axis: Axis index (1 .. 2).
        :argument steps: Number of steps to move relative to the current position.
            Positive values move in positive direction;
            negative values move in negative direction.
        :raises: QMI_UsageException by input steps number out of range.
        """
        if abs(steps) > 2**31:
            raise QMI_UsageException(f"Number of steps out of range {steps}")
        self.select_channel(channel)
        self._set_axis_attribute("PR", axis, steps)

    @rpc_method
    def stop(self, axis: int) -> None:
        """Stop current movement.

        :argument axis: Axis index (1 .. 2).
        :raises: QMI_UsageException by invalid axis index number.
        """
        if axis not in (1, 2):
            raise QMI_UsageException(f"Invalid axis index {axis}")
        self._write(f"{axis}ST")



#! /usr/bin/env python3

import sys

import qmi

from qmi.core.exceptions import QMI_Exception
from qmi.core.read_keyboard import KeyboardReader
from qmi.instruments.dummy.noisy_sine_generator import NoisySineGenerator

from noisy_sine_generator_controller import NoisySineGeneratorSettings, NoisySineGeneratorController

def main():
    qmi.start("nsg_service", "qmi.conf")

    try:
        # Make the instrument as a QMI instrument.
        nsg = qmi.make_instrument("nsg", NoisySineGenerator)

        # Create the task; the task initializer applies the settings.
        settings = NoisySineGeneratorSettings(frequency=2.0, amplitude=100.0, noise=1.0)
        controller = qmi.make_task("controller",
                                   NoisySineGeneratorController,
                                   generator=nsg,
                                   settings=settings,
                                   sample_time=1.0)

        # Start the task.
        controller.start()
        print("Service running {!r} - type 'Q' + Enter to terminate.".format(qmi.context().name))

        # Run until termination is requested, either by RPC call or by the user.
        kbd = KeyboardReader()
        while True:
            if qmi.context().wait_until_shutdown(duration=0.1):
                print("Context shutdown requested - stopping the service.")
                break
            if kbd.poll_quit():
                print("Shutdown requested by user - stopping the service.")
                break

        # Wait for the task threads to finish.
        controller.stop()
        controller.join()

    except QMI_Exception as e:
        print("ERROR: {}".format(e))
        return 1

    finally:
        # Always stop the context.
        qmi.stop()

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Instrument driver for the Toptica DLC Pro digital laser controller."""

import logging
import time
from typing import Any, List, Tuple

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport


# Module logger.
_logger = logging.getLogger(__name__)


class Toptica_DLC(QMI_Instrument):
    """Instrument driver for the Toptica DLC Pro digital laser controller.

    This driver implements a single connection to the instrument. The communication is using a Scheme-based command
    language on TCP port 1998;

    NOTE: the driver does not currently support the parameter monitoring facility on TCP port 1999.

    Parameters:
        context:    QMI context.
        name:       name for the instrument instance.
        transport:  transport string.
    """

    # Instrument should respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver."""
        super().__init__(context, name)
        self._transport = create_transport(transport)

    @rpc_method
    def open(self) -> None:
        """"Open connection with the instrument."""
        _logger.info("Opening connection to instrument")
        self._transport.open()
        try:
            self._main_handshake()
        except Exception:
            self._transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        """Close connection with the instrument."""
        _logger.info("Closing connection to instrument")
        super().close()
        self._transport.close()

    def _main_handshake(self) -> None:
        """Read initial welcome message from Telnet port."""
        self._transport.read_until(message_terminator=b"DeCoF Command Line\r\n> ", timeout=self.RESPONSE_TIMEOUT)

    @staticmethod
    def _parse_value_string(value_string: str) -> Any:
        """Parse a response string to a value."""
        # NOTE: for now, we don't handle ints, but instead return a float.
        if value_string == "#f":
            # Value is a boolean False.
            return False
        elif value_string == "#t":
            # Value is a boolean True.
            return True
        elif len(value_string) >= 2 and value_string.startswith('"') and value_string.endswith('"'):
            # Value is a string.
            return value_string[1:-1]
        elif value_string.startswith('(') and value_string.endswith(')'):
            # Tuple: this breaks for nested nested tuples.
            return tuple(Toptica_DLC._parse_value_string(x) for x in value_string[1:-1].split())
        else:
            # Assume it is a float or can be converted to a float.
            return float(value_string)

    @staticmethod
    def _value_to_string(value: Any) -> str:
        """Convert a value to its string representation."""
        # Only handles bools, ints, and floats.
        if isinstance(value, bool):
            return "#t" if value else "#f"
        elif isinstance(value, (int, float)):
            return repr(value)
        elif isinstance(value, str):
            return value
        else:
            raise ValueError(f"Invalid value or type for {value}")

    @rpc_method
    def get_timestamped_parameter(self, parameter_name: str) -> Tuple[float, Any]:
        """Get parameter value from the controller.

        Parameters:
            parameter_name: string identifier of the parameter to get.
        """
        # Send request.
        command = "(param-ref '{})\n".format(parameter_name).encode()
        self._transport.write(command)

        # Get response.
        response = self._transport.read_until(b"\n", timeout=self.RESPONSE_TIMEOUT)
        timestamp = time.time()

        # Read until prompt.
        self._transport.read_until(b"> ", timeout=self.RESPONSE_TIMEOUT)

        # Decode response.
        response = response[:-1]  # strip newline.
        decoded_response = response.decode()
        if decoded_response.startswith("Error:"):
            raise ValueError("While getting parameter {!r}: {!r}".format(parameter_name, decoded_response))

        return timestamp, self._parse_value_string(decoded_response)

    @rpc_method
    def get_parameter(self, parameter_name: str) -> Any:
        """Get a single parameter value from the controller.

        Returns a timestamp of when the parameteres were requested from the controller and the parameter value.

        Parameters:
            parameter_name: string identifier of the parameter to get.
        """
        _, value = self.get_timestamped_parameter(parameter_name)
        return value

    @rpc_method
    def get_multiple_timestamped_parameters(self, parameter_names: List[str]) -> Tuple[float, Tuple[Any, ...]]:
        """Get multiple parameter values from the controller.

        Returns a timestamp of when the parameters were requested from the controller and a tuple of parameter values.

        Note: this method cannot handle nested tuples (e.g. one of the parameter values is a tuple).

        Parameters:
            parameter_names:    a list of string identifiers of the parameters to get.
        """
        # Send request.
        command = "`({})\n".format(
            "".join(",(param-ref '{})".format(parameter_name) for parameter_name in parameter_names)).encode()
        self._transport.write(command)

        # Get response.
        response = self._transport.read_until(b"\n", timeout=self.RESPONSE_TIMEOUT)
        timestamp = time.time()

        # Read until prompt.
        self._transport.read_until(b"> ", timeout=self.RESPONSE_TIMEOUT)

        # Decode response.
        response = response[:-1]  # strip newline
        decoded_response = response.decode()
        if decoded_response.startswith("Error:"):
            raise ValueError("While getting multiple parameters {!r}: {!r}".format(parameter_names, response))

        # Check that a tuple was returned.
        retval = self._parse_value_string(decoded_response)
        assert isinstance(retval, tuple)

        return timestamp, retval

    @rpc_method
    def get_multiple_parameters(self, parameter_names: List[str]) -> Tuple[Any, ...]:
        """Get mulitple parameter values.

        Parameters:
            parameter_names:    a list of string identifiers of the parameters to get.
        """
        _, values = self.get_multiple_timestamped_parameters(parameter_names)
        return values

    @rpc_method
    def set_parameter(self, parameter_name: str, value: Any) -> float:
        """Set parameter value from the controller.

        Returns the time at which the set command was sent to the controller.

        Parameters:
            parameter_name: string identifier of the parameter to set.
            value:          new value for the parameter.
        """
        # Convert value to string respresentation.
        value_string = self._value_to_string(value)

        # Send command.
        command = "(param-set! '{} {})\n".format(parameter_name, value_string).encode()
        self._transport.write(command)
        timestamp = time.time()

        # Get response.
        response = self._transport.read_until(b"\n", self.RESPONSE_TIMEOUT)

        # Read until prompt.
        self._transport.read_until(b"> ", self.RESPONSE_TIMEOUT)

        # Decode response.
        response = response[:-1]  # strip newline
        decoded_response = response.decode()
        if decoded_response.startswith("Error:"):
            raise ValueError("while setting parameter {!r}: {!r}".format(parameter_name, response))

        return timestamp

    @rpc_method
    def set_multiple_parameters(self, parameters: List[Tuple[str, Any]]) -> float:
        """Set parameter value from the controller.

        Returns the time at which the set command was sent to the controller.

        Parameters:
            parameters: a list of tuples with parameter_name-value pairs.
        """
        # Compose command.
        subcommands = [
            "(param-set! '{} {})".format(pname, self._value_to_string(pvalue)) for pname, pvalue in parameters
        ]
        command = "(+ {})\n".format("".join(subcommands)).encode()

        # Send command.
        self._transport.write(command)
        timestamp = time.time()

        # Get response.
        response = self._transport.read_until(b"\n", self.RESPONSE_TIMEOUT)

        # Read until prompt.
        self._transport.read_until(b"> ", self.RESPONSE_TIMEOUT)

        # Decode response.
        response = response[:-1]  # strip newline.
        decoded_response = response.decode()
        if decoded_response.startswith("Error:"):
            raise ValueError("while setting multiple parameters {!r}: {!r}".format(
                [p for p, _ in parameters], response)
            )

        return timestamp




import threading
import datetime
from typing import Dict, List, Union

import numpy as np
import h5py

from qmi.core.thread import QMI_Thread


# Supported types for HDF5 attribute values.
_AttributeValueType = Union[int, float, str]


class _HDF5RecorderThread(QMI_Thread):

    def __init__(self, filename: str, write_interval: float, keep_open: bool) -> None:
        super().__init__()

        self._filename = filename
        self._write_interval = write_interval
        self._keep_open = keep_open
        self._recordings: Dict[str, List[np.ndarray]] = {}
        self._attributes: Dict[str, Dict[str, _AttributeValueType]] = {}
        self._condition = threading.Condition(threading.Lock())

    def run(self) -> None:

        HDF5_FILE_MODE = "a"  # open for read/write access, create file if it does not exist
        fo = None

        recordings: Dict[str, List[np.ndarray]] = {}
        pending_attributes: Dict[str, Dict[str, _AttributeValueType]] = {}
        new_attributes: Dict[str, Dict[str, _AttributeValueType]] = {}

        quitflag = False
        while not quitflag:

            with self._condition:
                while not (self._shutdown_requested or len(self._recordings) != 0):
                    # Wake up every once in a while to write collected data.
                    self._condition.wait(self._write_interval)
                # We're still holding the mutex here. Obtain safe-to-use references to shared variables.
                (recordings, self._recordings) = (self._recordings, recordings)
                (new_attributes, self._attributes) = (self._attributes, new_attributes)
                quitflag = self._shutdown_requested

            # The mutex is now relinquished. We can now process the recordings at a leasurely pace.

            for (dset_name, dset_values_list) in recordings.items():
                append_dset_values = np.concatenate(dset_values_list)
                n = len(append_dset_values)
                if n != 0:
                    # We have to write data...
                    if fo is None:
                        # Open for read/write (file must exist)
                        fo = h5py.File(self._filename, HDF5_FILE_MODE)

                    # Check if the dataset already exists
                    if dset_name in fo:
                        # Append data to existing dataset
                        dset = fo[dset_name]
                        old_size = len(dset)
                        new_size = old_size + n

                        dset.resize((new_size, ))
                        dset[-n:] = append_dset_values
                    else:
                        # Create a new dataset ; prepare for it to be extended.
                        dset = fo.create_dataset(dset_name, data=append_dset_values, maxshape=(None, ))

                    # Write any attributes that were queued for this dataset.
                    if dset_name in pending_attributes:
                        new_dset_attrs = pending_attributes.pop(dset_name)
                        dset.attrs.update(new_dset_attrs)
                    if dset_name in new_attributes:
                        new_dset_attrs = new_attributes.pop(dset_name)
                        dset.attrs.update(new_dset_attrs)

            # Loop over the newly queued attributes that have not yet been handled.
            for (dset_name, new_dset_attrs) in new_attributes.items():
                if dset_name in pending_attributes:
                    # We already know that this dataset does not yet exist.
                    pending_attributes[dset_name].update(new_dset_attrs)
                else:
                    # Open the HDF5 file if it is not yet open.
                    if fo is None:
                        fo = h5py.File(self._filename, HDF5_FILE_MODE)
                    if dset_name in fo:
                        # Write attributes to existing dataset.
                        fo[dset_name].attrs.update(new_dset_attrs)
                    else:
                        # This dataset does not exist yet.
                        # Hold on to the attributes; we will write them as soon as the dataset is created.
                        pending_attributes[dset_name] = new_dset_attrs

            new_attributes.clear()

            if not self._keep_open:
                if fo is not None:
                    fo.close()
                    fo = None

            recordings.clear()  # All recordings were processed.

        if fo is not None:
            fo.close()
            fo = None

    def _request_shutdown(self) -> None:
        with self._condition:
            self._condition.notify_all()

    def record(self, dset_name: str, dset_values: np.ndarray) -> None:
        # NOTE: Called from an outside thread.

        if len(dset_values) == 0:
            return

        dset_values = dset_values.copy()  # copy it.

        with self._condition:

            if dset_name not in self._recordings:
                self._recordings[dset_name] = [dset_values]
            else:
                self._recordings[dset_name].append(dset_values)

    def set_attribute(self, dset_name: str, attr_name: str, attr_val: _AttributeValueType) -> None:
        with self._condition:
            dset_attrs = self._attributes.setdefault(dset_name, {})
            dset_attrs[attr_name] = attr_val


class HDF5Recorder:

    DEFAULT_WRITE_INTERVAL = 60.0 # Default interval to write, in seconds.

    def __init__(self, filename: str, write_interval: float = DEFAULT_WRITE_INTERVAL, keep_open: bool = False):

        # Note, this is local time, so ambiguous w.r.t. summer/winter time.
        local_timestamp_string = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = filename.replace("<localtime>", local_timestamp_string)

        self._recorder_thread = _HDF5RecorderThread(filename, write_interval, keep_open)
        self._recorder_thread.start()

    def close(self) -> None:
        self._recorder_thread.shutdown()
        self._recorder_thread.join()
        self._recorder_thread = None  # type: ignore

    def record(self, dset_name: str, dset_values: np.ndarray) -> None:
        self._recorder_thread.record(dset_name, dset_values)

    def set_attribute(self, dset_name: str, attr_name: str, attr_val: Union[int, float, str]) -> None:
        """Add or update a HDF5 attribute in the dataset.

        Attributes can only be written to existing datasets.
        If the named dataset does not yet exist, the attribute will be cached in the `HDF5Recorder`
        until the dataset is created by recording data to it. If the dataset is never created,
        pending attribute values are lost when the `HDF5Recorder` is closed.

        Parameters:
            dset_name: Name of the dataset to write the attribute to.
            attr_name: Attribute name.
            attr_val: Attribute value. Only numbers and strings are currently supported.
        """
        self._recorder_thread.set_attribute(dset_name, attr_name, attr_val)



"""Module that defines packet formats for the UDP Responder.
"""

import enum
import ctypes
import logging
from typing import ClassVar

from qmi.core.exceptions import QMI_RuntimeException

_logger = logging.getLogger(__name__)

# The magic number to recognize QMI UDP packets.
MAGIC = 0x00494d51  # Magic value. This is stored little endian, and then reads 'QMI\0' in ASCII.


class QMI_LittleEndianStructure(ctypes.LittleEndianStructure):
    """The standard ctypes LittleEndianStructure, with a proper repr() function."""

    def __repr__(self) -> str:
        fstr_list = []
        for (fname, ftype, *_) in self._fields_:
            value = getattr(self, fname)
            if isinstance(value, (QMI_LittleEndianStructure, int, float, bytes)):
                vstr = repr(value)
            elif isinstance(value, ctypes.Array):
                vstr = "[{}]".format(", ".join(repr(e) for e in value))
            else:
                _logger.error("Unhandled type: %s %s", type(value), value)
                vstr = repr(value)
            fstr = "{}={}".format(fname, vstr)
            fstr_list.append(fstr)
        return "{}({})".format(self.__class__.__name__, ", ".join(fstr_list))


@enum.unique
class QMI_UdpResponderMessageTypeTag(enum.Enum):
    # Requests sent by qmitool or similar tools.
    CONTEXT_INFO_REQUEST  = 0x201 # Broadcast or targeted.
    CONTEXT_KILL_REQUEST  = 0x202 # Broadcast or targeted.
    # Packets sent by the Contexts.
    CONTEXT_INFO_RESPONSE = 0x101 #
    CONTEXT_STARTUP       = 0x102 # Broadcast.
    CONTEXT_SHUTDOWN      = 0x103 # Broadcast.


class QMI_UdpResponderPacketHeader(QMI_LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('magic'         , ctypes.c_uint32),  # Value: 0x00494d51 ('QMI\0').
        ('pkt_type_tag'  , ctypes.c_uint16),  # Entry from QMI_UdpResponderMessageTypeTag.
        ('pkt_id'        , ctypes.c_uint64),  # This message's ID.
        ('pkt_timestamp' , ctypes.c_double),  # This message's POSIX timestamp.
    ]

class QMI_UdpResponderContextDescriptor(QMI_LittleEndianStructure):
    _pack_ = 1
    _fields_ = [
        ('pid'           , ctypes.c_int32    ),  # PID of the process that owns the QMI_Context.
        ('name'          , ctypes.c_char * 64),  # Name of the context
        ('workgroup_name', ctypes.c_char * 64),
        ('port'          , ctypes.c_int32    )   # Listening TCP port of the context. (-1) means: no port.
    ]

# Concrete packets follow below:

class QMI_UdpResponderContextInfoRequestPacket(QMI_UdpResponderPacketHeader):
    _pack_ = 1
    _fields_ = [
        ('workgroup_name_filter', ctypes.c_char * 64),
        ('context_name_filter', ctypes.c_char * 64)
    ]  # No extra fields.

    @staticmethod
    def create(
        pkt_id: int,
        pkt_timestamp: float,
        workgroup_name_filter: bytes,
        context_name_filter: bytes
    ) -> 'QMI_UdpResponderContextInfoRequestPacket':
        return QMI_UdpResponderContextInfoRequestPacket(
            MAGIC,
            QMI_UdpResponderMessageTypeTag.CONTEXT_INFO_REQUEST.value,
            pkt_id,
            pkt_timestamp,
            workgroup_name_filter,
            context_name_filter
        )


class QMI_UdpResponderContextInfoResponsePacket(QMI_UdpResponderPacketHeader):
    _pack_ = 1
    _fields_ = [
        ('request_pkt_id'        , ctypes.c_uint64                  ),
        ('request_pkt_timestamp' , ctypes.c_double                  ),
        ('context'               , QMI_UdpResponderContextDescriptor)
    ]

    @staticmethod
    def create(
        pkt_id: int,
        pkt_timestamp: float,
        request_pkt_id: int,
        request_pkt_timestamp: float,
        context_pid: int,
        context_name: bytes,
        workgroup_name: bytes,
        context_port: int
    ) -> 'QMI_UdpResponderContextInfoResponsePacket':
        context_descriptor = QMI_UdpResponderContextDescriptor(context_pid, context_name, workgroup_name, context_port)
        return QMI_UdpResponderContextInfoResponsePacket(
            MAGIC, QMI_UdpResponderMessageTypeTag.CONTEXT_INFO_RESPONSE.value,
            pkt_id,
            pkt_timestamp,
            request_pkt_id,
            request_pkt_timestamp,
            context_descriptor
        )


class QMI_UdpResponderKillRequestPacket(QMI_UdpResponderPacketHeader):
    _pack_ = 1
    _fields_: ClassVar[list] = []  # No extra fields.

    @staticmethod
    def create(pkt_id: int, pkt_timestamp: float) -> 'QMI_UdpResponderKillRequestPacket':
        return QMI_UdpResponderKillRequestPacket(
            MAGIC,
            QMI_UdpResponderMessageTypeTag.CONTEXT_KILL_REQUEST.value,
            pkt_id,
            pkt_timestamp
        )


_packet_type_lookup = {
    QMI_UdpResponderMessageTypeTag.CONTEXT_INFO_REQUEST  : QMI_UdpResponderContextInfoRequestPacket,
    QMI_UdpResponderMessageTypeTag.CONTEXT_KILL_REQUEST  : QMI_UdpResponderKillRequestPacket,
    QMI_UdpResponderMessageTypeTag.CONTEXT_INFO_RESPONSE : QMI_UdpResponderContextInfoResponsePacket
}


def unpack_qmi_udp_packet(packet: bytes) -> QMI_UdpResponderPacketHeader:

    packet_size = len(packet)

    if packet_size < ctypes.sizeof(QMI_UdpResponderPacketHeader):
        raise QMI_RuntimeException("Bad QMI UDP packet: too short.")

    header = QMI_UdpResponderPacketHeader.from_buffer_copy(packet)

    if header.magic != MAGIC:
        raise QMI_RuntimeException("Bad QMI UDP packet: bad magic field.")

    packet_type_tag = QMI_UdpResponderMessageTypeTag(header.pkt_type_tag)

    packet_type = _packet_type_lookup.get(packet_type_tag)
    if packet_type is None:
        raise QMI_RuntimeException("Bad QMI UDP packet: unknown packet type tag.")

    expected_packet_size = ctypes.sizeof(packet_type)

    if packet_size != expected_packet_size:
        raise QMI_RuntimeException(
            "Bad UDP packet: unexpected size (tag = {}, actual = {}, expected = {})".format(
                packet_type_tag, packet_size, expected_packet_size
            )
        )

    return packet_type.from_buffer_copy(packet)



"""
Quantum Opus, low-noise amplifier module.

The qmi.instruments.quantum_opus package provides support for:
- QO-AMP-SIM module.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.quantum_opus.amp_sim_module import AmpSimModule as QuantumOpus_QoAmpSim



"""Instrument driver for the Montana Cryostation S50."""

from dataclasses import dataclass
from enum import Enum
from ipaddress import IPv4Address
import json
import logging
from typing import Dict, Optional, Union, cast
import http
import urllib.error
import urllib.parse
import urllib.request

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.exceptions import QMI_InstrumentException, QMI_RuntimeException
from qmi.core.rpc import rpc_method

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@dataclass
class Montana_CryostationS50_Thermometer_Properties:
    """
    Dataclass for the Montana S50 thermometer properties of the sample.
    """
    resistance_ok: bool
    resistance: float
    temperature_ok: bool
    temperature: float
    temperature_stability: float
    temperature_stability_ok: bool
    temperature_stable: bool


class Montana_CryostationS50_System_State(Enum):
    """
    Possible states that the system can be in.
    """
    INVALID = "Invalid"
    IDLE = "Idle"
    CONFIGURING = "Configuring"
    READY = "Ready"
    WAITING_FOR_VACUUM_SYSTEM = "WaitingForVacuumSystem"
    CLEANING_VACUUM_LINES = "CleaningVacuumLines"
    CHECKING_FOR_LEAKS = "CheckingForLeaks"
    BAKEOUT_PREHEATING = "BakeoutPreheating"
    BAKING_OUT = "BakingOut"
    PURGING = "Purging"
    PULLING_VACUUM = "Pulling Vacuum"
    COOLING_DOWN = "CoolingDown"
    ACQUIRING_TARGET = "AcquiringTarget"
    STABLE_AT_TARGET = "StableAtTarget"
    WARMING_UP = "WarmingUp"
    VENTING = "Venting"


class Montana_CryostationS50_System_Goal(Enum):
    """
    Possible goals of the system.
    """
    NONE = "None"
    PULL_VACUUM = "PullVacuum"
    VENT = "Vent"
    COOLDOWN = "Cooldown"
    WARMUP = "Warmup"


class Montana_CryostationS50(QMI_Instrument):
    """
    Instrument driver for the Montana Cryostation S50.
    The S50 uses a RESTful API for communication.
    """

    # Response Timeout.
    RESPONSE_TIMEOUT = 5.0

    # Subsystem strings.
    CONTROLLER_SUBSYSTEM_STR = "controller"
    CRYOCOOLER_SUBSYSTEM_STR = "cooler/cryocooler"
    SAMPLE_CHAMBER_TEMP_CONTROLLER_SUBSYSTEM_STR = "sampleChamber/temperatureControllers/platform"

    # System states.
    SYSTEM_STATES = {value.value: value for value in Montana_CryostationS50_System_State}

    # System goals.
    SYSTEM_GOALS = {value.value: value for value in Montana_CryostationS50_System_Goal}

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 ipv4_address: str
                 ) -> None:
        """
        Initializes a Montana_CryostationS50 instance.

        Note:
            This method should not be called directly by a QMI user.
            Always instantiate an instrument via a QMI_Context `make_instrument()` call.

        Arguments:
            context:        the QMI context
            name:           the name of the instrument instance
            ipv4_address:   the IP address of the cryostation
        """

        super().__init__(context, name)
        self._host = IPv4Address(ipv4_address)  # raises ValueError for improper address
        self._base_url = f"http://{self._host}:47101/v1"
        self._timeout = Montana_CryostationS50.RESPONSE_TIMEOUT
        self._is_open = False

    def _send_request(self, url: str, request_type: str, data: Optional[bytes] = None) -> Optional[Dict[str, str]]:
        """
        Send a GET, PUT or POST request to the instrument.

        Parameters:
            url:            the url to send the request to.
            request_type:   the type of the request.
            data:           optional data for PUT and POST requests.
        """
        valid_request_types = ["GET", "PUT", "POST"]
        if not self._is_open:
            raise QMI_RuntimeException("Open device first")
        _logger.debug("Request: %s", url)

        if request_type not in valid_request_types:
            raise QMI_RuntimeException(f"Valid request types are {valid_request_types}")

        # send request and check for response
        req = urllib.request.Request(url, method=request_type, data=data)
        try:
            with urllib.request.urlopen(req) as response:
                status = response.status
                if status != http.HTTPStatus.OK:
                    raise QMI_InstrumentException(f"HTTP error code: {status}")

                if request_type == 'GET':
                    return json.loads(response.read())

        except urllib.error.HTTPError as exc:
            raise QMI_InstrumentException("Error in communication with device") from exc

        return None

    def _set_property(self, subsystem: str, key: str, value: Union[str, float, bool]) -> None:
        """
        Send a put request to set the property.

        Parameters:
            subsystem:  the subsystem to set the property for.
            key:        the name of the property to set.
            value:      the value to set.
        """
        # create json dictionary and encode for PUT
        data = {
            key: value
        }
        json_data = json.dumps(data, indent=4)
        encoded_data = bytes(json_data, 'utf-8')

        url = f"{self._base_url}/{subsystem}/properties/{key}"
        self._send_request(url, 'PUT', encoded_data)

    def _get_property(self, subsystem: str, key: str) -> str:
        """
        Send a get request to get the property.

        Parameters:
            subsystem:  the subsystem to get the property for.
            key:        the name of the property to get.
        """
        url = f"{self._base_url}/{subsystem}/properties/{key}"
        response = self._send_request(url, 'GET')
        if response is None:
            raise QMI_InstrumentException("Did not receive any response from a GET request")
        return response[key]

    def _run_method(self, subsystem: str, key: str) -> None:
        url = f"{self._base_url}/{subsystem}/methods/{key}()"
        self._send_request(url, 'POST')

    def _set_platform_bakeout(self, enable: bool) -> None:
        self._set_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutEnabled", enable)

    @rpc_method
    def open(self) -> None:
        """Opens a connection to the Cryostation instrument."""
        _logger.info("Opening connection to %s", self._name)
        super().open()

    @rpc_method
    def close(self) -> None:
        """Closes a connection to the Cryostation instrument."""
        self._check_is_open()
        _logger.info("Closing connection to %s", self._name)
        super().close()

    @rpc_method
    def get_system_goal(self) -> Montana_CryostationS50_System_Goal:
        """
        Get the current goal of the system.

        Returns:
            the goal of the system as an enum.
        """
        _logger.info("Getting system goal of %s", self._name)
        goal = self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "systemGoal")
        return self.SYSTEM_GOALS[goal]

    @rpc_method
    def get_system_state(self) -> Montana_CryostationS50_System_State:
        """
        Get the current state of the system.

        Returns:
            the state of the system as an enum.
        """
        _logger.info("Gettng system state of %s", self._name)
        state = self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "systemState")
        return self.SYSTEM_STATES[state]

    @rpc_method
    def cooldown_allowed(self) -> bool:
        """
        Check if the system is in a state where a cooldown can be started.

        Returns:
            true if the system can be cooled down.
        """
        _logger.info("Checking if cooldown is allowed for %s", self._name)
        return bool(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "canCooldown"))

    @rpc_method
    def warmup_allowed(self) -> bool:
        """
        Check if the system is in a state where a warmup can be started.

        Returns:
            true if the system can be warmed up down.
        """
        _logger.info("Checking if warmup is allowed for %s", self._name)
        return bool(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "canWarmup"))

    @rpc_method
    def set_platform_target_temperature(self, temp: float) -> None:
        """
        Set the platform target temperature for cooldown.

        Parameters:
            temp: temperature to set in Kelvin.
        """
        _logger.info("Setting platform target temperature of %s", self._name)
        self._set_property(self.CONTROLLER_SUBSYSTEM_STR, "platformTargetTemperature", temp)

    @rpc_method
    def get_platform_target_temperature(self) -> float:
        """
        Get the temperature that the platform is to be cooled down to.

        Returns:
            temperature in Kelvin.
        """
        _logger.info("Getting platform target temperature of %s", self._name)
        return float(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "platformTargetTemperature"))

    @rpc_method
    def get_platform_target_temperature_limit(self) -> float:
        """
        Get the temperature limit that the platform can be cooled down to.

        Returns:
            temperature in Kelvin
        """
        _logger.info("Getting max platform temperature for %s", self._name)
        return float(self._get_property(self.SAMPLE_CHAMBER_TEMP_CONTROLLER_SUBSYSTEM_STR, "temperatureLimit"))

    @rpc_method
    def start_cooldown(self) -> None:
        """
        Cool down the sample chamber to the target temperature.
        """
        _logger.info("Starting cooldown of %s", self._name)
        self._run_method(self.CONTROLLER_SUBSYSTEM_STR, "cooldown")

    @rpc_method
    def start_warmup(self) -> None:
        """
        Warm up the sample chamber.
        """
        _logger.info("Starting warmup of %s", self._name)
        self._run_method(self.CONTROLLER_SUBSYSTEM_STR, "warmup")

    @rpc_method
    def enable_platform_bakeout(self) -> None:
        """
        Enable bakeout of platform when cooling down or pulling vacuum.
        """
        _logger.info("Enabling platform bakeout of %s", self._name)
        self._set_platform_bakeout(True)

    @rpc_method
    def disable_platform_bakeout(self) -> None:
        """
        Disable bakeout of platform when cooling down or pulling vacuum.
        """
        _logger.info("Disabling platform bakeout of %s", self._name)
        self._set_platform_bakeout(False)

    @rpc_method
    def is_platform_bakeout_enabled(self) -> bool:
        """
        Check if the platform bakeout procedure is enabled.

        Returns:
            true if bakeout is enabled else False.
        """
        _logger.info("Checking is platform bakeout is enable for %s", self._name)
        return bool(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutEnabled"))

    @rpc_method
    def get_platform_bakeout_temperature_limit(self) -> float:
        """
        Get the temperature limit that the platform can be baked to.

        Returns:
            temperature in Kelvin.
        """
        _logger.info("Getting max platform bakeout temperature for %s", self._name)
        return float(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutTemperatureLimit"))

    @rpc_method
    def set_platform_bakeout_temperature(self, temp: float) -> None:
        """
        Set the platform bakeout temperature.

        Parameters:
            temp: temperature to set in Kelvin.
        """
        _logger.info("Setting platform bakeout temperature for %s", self._name)
        self._set_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutTemperature", temp)

    @rpc_method
    def get_platform_bakeout_temperature(self) -> float:
        """
        Get the platform bakeout temperature.

        Returns:
            temperature in Kelvin.
        """
        _logger.info("Getting platform bakeout temperature for %s", self._name)
        return float(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutTemperature"))

    @rpc_method
    def set_platform_bakeout_time(self, time: float) -> None:
        """
        Set the platform bakeout time.

        Parameters:
            time: time to set in seconds.
        """
        _logger.info("Setting platform bakeout time for %s", self._name)
        self._set_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutTime", time)

    @rpc_method
    def get_platform_bakeout_time(self) -> float:
        """
        Get the platform bakeout time.

        Returns:
            time in seconds.
        """
        _logger.info("Getting platform bakeout time for %s", self._name)
        return float(self._get_property(self.CONTROLLER_SUBSYSTEM_STR, "platformBakeoutTime"))

    @rpc_method
    def is_cryocooler_running(self) -> bool:
        """
        Check if the cryocooler is running.

        Returns:
            true if the cryocooler is running else False.
        """
        _logger.info("Checking if cryocooler is running for %s", self._name)
        return bool(self._get_property(self.CRYOCOOLER_SUBSYSTEM_STR, "cryocoolerRunning"))

    @rpc_method
    def is_cryocooler_connected(self) -> bool:
        """
        Check if the cryocooler is connected and communicating properly.

        Returns:
            true if the cryocooler is connected and communicating properly else False.
        """
        _logger.info("Checking if cryocooler is connected for %s", self._name)
        return bool(self._get_property(self.CRYOCOOLER_SUBSYSTEM_STR, "deviceConnected"))

    @rpc_method
    def get_sample_thermometer_properties(self) -> Montana_CryostationS50_Thermometer_Properties:
        """
        Get the thermometer properties of the sample.

        Returns:
            Dataclass with the properties.
        """
        _logger.info("Getting thermometer properties for sample for %s", self._name)
        dict_data: Dict[str, str] = cast(Dict[str, str], self._get_property(
            self.SAMPLE_CHAMBER_TEMP_CONTROLLER_SUBSYSTEM_STR + "/thermometer", "sample"))
        return Montana_CryostationS50_Thermometer_Properties(
            bool(dict_data["resistanceOK"]),
            float(dict_data["resistance"]),
            bool(dict_data["temperatureOK"]),
            float(dict_data["temperature"]),
            float(dict_data["temperatureStability"]),
            bool(dict_data["temperatureStabilityOK"]),
            bool(dict_data["temperatureStable"]))
